@@ -46,9 +46,6 @@
 
 #include "qdebug.h"
 #include <QtCore/qhashfunctions.h>
-#ifndef QT_NO_REGEXP
-# include "qregexp.h"
-#endif
 #ifndef QT_NO_DATASTREAM
 # include "qdatastream.h"
 #endif
@@ -1227,10 +1224,12 @@ QString QKeySequence::encodeString(int key)
 static inline void addKey(QString &str, const QString &theKey, QKeySequence::SequenceFormat format)
 {
     if (!str.isEmpty()) {
-        if (format == QKeySequence::NativeText)
+        if (format == QKeySequence::NativeText) {
+            //: Key separator in shortcut string
             str += QCoreApplication::translate("QShortcut", "+");
-        else
+        } else {
             str += QLatin1Char('+');
+        }
     }
 
     str += theKey;
@@ -1289,7 +1288,7 @@ QString QKeySequencePrivate::encodeString(int key, QKeySequence::SequenceFormat
 
     QString p = keyName(key, format);
 
-#if defined(Q_OS_OSX)
+#if defined(Q_OS_MACOS)
     if (nativeText)
         s += p;
     else
@@ -1405,7 +1404,7 @@ QKeySequence::SequenceMatch QKeySequence::matches(const QKeySequence &seq) const
 */
 QKeySequence::operator QVariant() const
 {
-    return QVariant(QVariant::KeySequence, this);
+    return QVariant(QMetaType::QKeySequence, this);
 }
 
 /*! \fn QKeySequence::operator int () const
@@ -1470,7 +1469,7 @@ bool QKeySequence::operator==(const QKeySequence &other) const
     Calculates the hash value of \a key, using
     \a seed to seed the calculation.
 */
-uint qHash(const QKeySequence &key, uint seed) Q_DECL_NOTHROW
+uint qHash(const QKeySequence &key, uint seed) noexcept
 {
     return qHashRange(key.d->key, key.d->key + QKeySequencePrivate::MaxKeyCount, seed);
 }
@@ -1525,7 +1524,7 @@ bool QKeySequence::operator< (const QKeySequence &other) const
 */
 bool QKeySequence::isDetached() const
 {
-    return d->ref.load() == 1;
+    return d->ref.loadRelaxed() == 1;
 }
 
 /*!
@@ -1694,3 +1693,5 @@ QDebug operator<<(QDebug dbg, const QKeySequence &p)
 */
 
 QT_END_NAMESPACE
+
+#include "moc_qkeysequence.cpp"
